module("L_BoschSolutionAlarmPanel1", package.seeall)

local ssl
local peer
local log = luup.log
local params = {
    mode = "client",
	protocol = "tlsv1",
    verify = {"none"},
	options = {"no_compression"},
	ciphers = "ALL",
}

-- Flags
local DEBUG_MODE = false
local THERMOSTATS_INITIALIZED = false
local FLAGS = {
	PARTITIONS = false,
	ZONES = true
}

-- Constants
local DEVICE_FILES = {
	BINARY_LIGHT       = "D_BoschSolutionOutput.xml",
	DOOR_SENSOR        = "D_DoorSensor1.xml",
	TEMPERATURE_SENSOR = "D_TemperatureSensor1.xml",
	HUMIDITY_SENSOR    = "D_HumiditySensor1.xml",
	LIGHT_SENSOR       = "D_LightSensor1.xml",
	MOTION_SENSOR      = "D_MotionSensor1.xml"
}

local DEVICE_TYPES = {
	BINARY_LIGHT       = "urn:schemas-upnp-org:device:BinaryLight:1",
	DOOR_SENSOR        = "urn:schemas-micasaverde-com:device:DoorSensor:1",
	TEMPERATURE_SENSOR = "urn:schemas-micasaverde-com:device:TemperatureSensor:1",
	HUMIDITY_SENSOR    = "urn:schemas-micasaverde-com:device:HumiditySensor:1",
	LIGHT_SENSOR       = "urn:schemas-micasaverde-com:device:LightSensor:1",
	MOTION_SENSOR      = "urn:schemas-micasaverde-com:device:MotionSensor:1"
}

local SID = {
	BOSCH      		= "urn:micasaverde-com:serviceId:BoschSolutionAlarmPanel1",
	PARTITION  		= "urn:micasaverde-com:serviceId:AlarmPartition2",
	SES        		= "urn:micasaverde-com:serviceId:SecuritySensor1",
	TEMP			= "urn:upnp-org:serviceId:TemperatureSensor1",
	HVACO			= "urn:upnp-org:serviceId:HVAC_UserOperatingMode1",
	HVACF 			= "urn:upnp-org:serviceId:HVAC_FanOperatingMode1",
	HVACHEAT 		= "urn:upnp-org:serviceId:TemperatureSetpoint1_Heat",
	HVACCOOL 		= "urn:upnp-org:serviceId:TemperatureSetpoint1_Cool",
	HADEVICE 		= "urn:micasaverde-com:serviceId:HaDevice1",
	ZWN_SID 		= "urn:micasaverde-com:serviceId:ZWaveNetwork1",
	SWP_SID 		= "urn:upnp-org:serviceId:SwitchPower1"
}

local TASK = {
	ERROR       = 2,
	ERROR_ALARM = -2,
	ERROR_STOP  = -4,
	SUCCESS     = 4,
	BUSY        = 1
}
-- LastUpdate
local DISPLAY_SECONDS = 20
local POLLING_RATE = 20 -- 5 -- 30 --"PollFrequency" | seconds

-- Globals
local numberOfAreas = 0
local numberOfPoints = 0
local numberOfOutputs = 0
local numberOfUsers = 0
local numberOfKeypads = 0
local numberOfDoors = 0
local numberOfEventRecordSize = 0
local historyMin = 0
local historyMax = 0
local lug_device = nil
local g_appendPtr -- The pointer passed to luup.chdev.append
local g_pinCodes = {}
local g_taskHandle = -1
local g_errorMessage = nil
local g_lastTask = os.time() -- The time when the status message was last updated.
local g_users = nil
local g_areas = {
	-- .panelID = ""
	-- .name = ""
	-- .devId
}
local g_points = {
	-- .panelID = ""
	-- .name = ""
}
local g_doors = {}
local g_outputs = {}
local g_passcode = ""
local ipAddress
local ipPort
local flagAlarm = false
local flagAlert = false
local historyLog = ""
local flagPasscode = false
---------------------------------------------------------
-----------------Generic Utils---------------------------
---------------------------------------------------------
local function log(text, level)
	luup.log("BoschSolutionPlugins::"..text, (level or 50))
end

local function debug(text)
	if (DEBUG_MODE) then
		luup.log("BoschSolutionPlugins::" .. text)
	end
end

local function warning(stuff)
	luup.log("BoschSolutionPlugins::warning: " .. stuff, 2)
end

local function error(stuff)
	luup.log("BoschSolutionPlugins::error: " .. stuff, 1)
end


local function displayMessage (text, mode)
	if (mode == TASK.ERROR_ALARM or mode == TASK.ERROR_STOP) then
		luup.task(text, TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
		if (mode == TASK.ERROR_STOP) then
			luup.set_failure(true, lug_device)
		end
		return
	end
	luup.task(text, mode, "Bosch Solution Plugin", g_taskHandle)
	-- Set message timeout.
	g_lastTask = os.time()
	luup.call_delay("clearTask", DISPLAY_SECONDS)
end

local hex2binArray = {
	["0"] = "0000",
	["1"] = "0001",
	["2"] = "0010",
	["3"] = "0011",
	["4"] = "0100",
	["5"] = "0101",
	["6"] = "0110",
	["7"] = "0111",
	["8"] = "1000",
	["9"] = "1001",
	["a"] = "1010",
	["b"] = "1011",
	["c"] = "1100",
	["d"] = "1101",
	["e"] = "1110",
	["f"] = "1111"
}

local bin2hexArray = {
    ["0000"] = "0",
    ["0001"] = "1",
    ["0010"] = "2",
    ["0011"] = "3",
    ["0100"] = "4",
    ["0101"] = "5",
    ["0110"] = "6",
    ["0111"] = "7",
    ["1000"] = "8",
    ["1001"] = "9",
    ["1010"] = "A",
    ["1011"] = "B",
    ["1100"] = "C",
    ["1101"] = "D",
    ["1110"] = "E",
    ["1111"] = "F"
}
local function Hex2Bin(s)
	local ret = ""
	local i = 0
	for i in string.gmatch(s, ".") do
		i = string.lower(i)
		ret = ret..hex2binArray[i]
	end
	return ret
end

local function Bin2Hex(s)
	local l = 0
	local h = ""
	local b = ""
	local rem
	l = string.len(s)
	rem = l % 4
	l = l-1
	h = ""
    -- need to prepend zeros to eliminate mod 4
    if (rem > 0) then
        s = string.rep("0", 4 - rem)..s
    end
    for i = 1, l, 4 do
        b = string.sub(s, i, i+3)
        h = h..bin2hexArray[b]
    end
    return h
end


local function ToHexNumber (n)
	if type(n) == "number" then
		return string.format( "%02X", n )
	end
end

function fromhex(str)
    if str then
		return (str:gsub('..', function (cc)
			return string.char(tonumber(cc, 16))
		end))
	else
		return "error"
	end
end

local function tohex(str)
    if str then
		return (str:gsub('.', function (c)
			return string.format('%02X', string.byte(c))
		end))
	else
		return "error"
	end
end

local function fromhex2(str)
    if str then
		return (str:gsub('.', function (cc)
			return string.format('%02d', string.byte(cc)) .. "|"--return string.byte(cc) .. "|"
		end))
	else
		return "error"
	end
end

local function parse(str,omit)
	local array = {}
	local parseArray = {}
	local countParseArray = 1
	local count = 1
	local countOmit = 1
	local indexOmit = 1
	local substring = ""
	for word in string.gmatch(str, ".") do
			array[count] = word
			count = count + 1
	end
	for i = 1, count - 1 do
		substring = substring .. array[i]
		if #omit >= 2 then
			if i < omit[indexOmit]  then
				if substring:len() == 2 then
					parseArray[countParseArray] = substring;
					substring = ""
					countParseArray = countParseArray + 1;
				end
			elseif i >= omit[indexOmit] and i <= omit[indexOmit+1] then
				if substring:len() == 4 then
					parseArray[countParseArray] = substring;
					substring = ""
					countParseArray = countParseArray + 1;
				end
				if i == omit[indexOmit + 1] then
					if indexOmit + 2 <  #omit then
						indexOmit = indexOmit + 2
					end
				end
			else
				if substring:len() == 2 then
					parseArray[countParseArray] = substring;
					substring = ""
					countParseArray = countParseArray + 1;
				end
			end
		else
			if substring:len() == 2 then
				parseArray[countParseArray] = substring;
				substring = ""
				countParseArray = countParseArray + 1;
			end
		end
	end
	return parseArray
end

local function wait(peer, err)
   if err == "timeout" or err == "wantread" then
		--log("***WANTREAD***")
		socket.select({peer}, nil,0)
    elseif err == "wantwrite" then
		--log("***WANTWRITE***")
		socket.select(nil, {peer},0)
    else
		--peer:close()
		--os.exit(1)
    end
end

local function send(str, peer, valid)
	local match = {"0","1","2","3","4","5","6","7","8","9"}
	local length  = tostring(str:len())
	local tosend = fromhex("01")

	for _,v in pairs(match) do
		if v == length then
			length = string.format('%02s',length)
		end
	end
	
	if valid then
		debug("(send) : Valid...")
		tosend = tosend .. fromhex(tostring(string.format( "%02X", length))) .. str
	else
		tosend = tosend .. fromhex(length) .. str
	end
	
	debug("(send) : Sending...")
	peer:settimeout(0.3)
	local succ, err = peer:send(tosend)
	debug("(send) : Waiting...")
	wait(peer, err)
	if succ then
		debug("(send) : Succesifully sent : [".. tohex(tosend) .. "] with length = [" .. succ .. "]")
	else
		debug("(send) : Sending error!!! .. Reconnecting ... ")
		--reconnect()
		if err then
			debug("(send) : ERROR : " .. err)
		end
	end
end

local function receive()
	local data = nil
	local status
	local partials = nil
	local length = 0
	local start = os.time()
	local stop
	while (not partials) or (tostring(partials)  == "")  do
		data , status, partials = peer:receive()
		stop = os.time()
		if (stop - start) > 10 then
			--debug("partial = [" .. tostring(partial) .. "]")
			--debug("status = [" .. tostring(status) .. "]")
			--debug("data = [" .. tostring(data) .. "]")
			return
		end
	end

	return partials
end

local function getPackageVersion()
	debug("(getPackageVersion) : Verifying packages...")
	local REQUIRED_VERSION = { '1', '0', '1', 'e', '1' }
	local stdout = io.popen("opkg list-installed | awk '/openssl-util/ {print $3}'")
	local version = stdout:read("*a")
	version = version:match("([^%s]+)")
	stdout:close()
	local i = 1
	for c in version:gmatch("[^.-]") do
			if c < REQUIRED_VERSION[i] then
				os.execute ("opkg install 'http://downloads.openwrt.org/attitude_adjustment/12.09/ramips/rt3883/packages/openssl-util_1.0.1e-1_ramips.ipk'")
				debug("(getPackageVersion) : Updating openssl-util!!!")
			end
			i = i + 1
	end
	
	local gtype = ""
	stdout = io.popen("cat /proc/cpuinfo | egrep machine")
	gtype = stdout:read("*a")
	stdout:close()
	
	if gtype:match("Sercomm%sG450") then
		REQUIRED_VERSION = { '0', '5', '1', '1', '0'}
	else
		REQUIRED_VERSION = { '0', '5', '5'}
	end
	stdout = io.popen("opkg list-installed | awk '/luasec/ {print $3}'")
	version = stdout:read("*a")
	version = version:match("([^%s]+)")
	stdout:close()
	i = 1
	for c in version:gmatch("[^.-]") do
			if c < REQUIRED_VERSION[i] then
				os.execute ("opkg update")
				os.execute ("opkg install luasec")
				debug("(getPackageVersion) : Updating luasec!!!")
			end
			i = i + 1
	end
	
	REQUIRED_VERSION = { '1', '0', '1', 'g', '1' }
	stdout = io.popen("opkg list-installed | awk '/libopenssl/ {print $3}'")
	version = stdout:read("*a")
	version = version:match("([^%s]+)")
	stdout:close()
	i = 1
	for c in version:gmatch("[^.-]") do
		if c < REQUIRED_VERSION[i] then
			os.execute ("opkg install 'http://downloads.openwrt.org/attitude_adjustment/12.09/ramips/rt3883/packages/libopenssl_1.0.1g-1_ramips.ipk'")
			debug("(getPackageVersion) : Updating libopenssl!!!")
		end
		i = i + 1
	end
	
	require("socket")
	ssl = require("ssl")
	peer = socket.tcp()
	debug("(getPackageVersion) : Packages OK")
end

local function UrlEncode (s)
	s = s:gsub("\n", "\r\n")
	s = s:gsub("([^%w])", function (c)
							  return string.format("%%%02X", string.byte(c))
						  end)
	return s
end
---------------------------------------------------------
---------------System Command Group----------------------
---------------------------------------------------------
local NAK = {
	["00"] 	= "Non-specific error",
	["01"] 	= "Checksum failure (UDP connections only)",
	["02"] 	= "Invalid size / length",
	["03"] 	= "Invalid command",
	["04"] 	= "Invalid interface state",
	["05"] 	= "Data out of range",
	["06"] 	= "No authority",
	["07"] 	= "Unsupported command",
	["08"]  = "Cannot Arm Area"
--	["Other condition"] 	= "Defined by the command",
}

local panelType = {
	["A8"]  = "B3512",
	["A0"] 	= "B4512",
	["A4"] 	= "B5512"
}

local function WhatAreYou()
	local tosend = fromhex("01")
	local receiveValue = ""
	send(tosend,peer)
	local receiveValue = ""
	receiveValue = receive()
	if receiveValue then
		debug ("*** WhatAreYou? = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** WhatAreYou? receive ERROR!" )
		return nil
	end
end

local function PasscodeAndUserTypeCheck(passcode)
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("06") .. fromhex("00") .. passcode .. fromhex("00")
	send(tosend,peer, true)
	receiveValue = receive()
	if receiveValue then
		debug ("*** PasscodeAndUserTypeCheck = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** PasscodeAndUserTypeCheck receive ERROR! ")
		return nil
	end
end

local function PincodeCheck(pincode)
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("3E") .. fromhex(pincode)
	send(tosend,peer, true)
	receiveValue = receive()
	if receiveValue then
		debug ("*** PincodeCheck = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** PincodeCheck receive ERROR! ")
		return nil
	end
end

local function RequestTextHistoryEvents(numberOfEvents, startingEvent) -- number 1bytes; startingEvent 4bytes
--[[
Note that due to internal size restrictions in the control panel, the current implementation only allows one
event to be returned at once. This restriction may be removed in future versions, so this should be taken
into account when designing an interface.
]]
	local tosend = ""
	local receiveValue = ""
	numberOfEvents = "01" -- delete this raw in the future when multiple events are supported
	tosend = fromhex("16") .. fromhex(numberOfEvents) .. fromhex(startingEvent)
	--tosend = fromhex("15") .. fromhex(numberOfEvents) .. fromhex(startingEvent)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestTextHistoryEvents = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestTextHistoryEvents receive ERROR! ")
		return nil
	end
end

local function RequestRawHistoryEvents(noEvents, startingEvent)
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("15") .. fromhex(noEvents) .. fromhex(startingEvent)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		log ("*** RequestRawHistoryEvents = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		return nil
	end
end

local function ResetSensors(area) -- multiple values allowed 2 bytes {"0001","0002"}
	local tosend = fromhex("18")
	local receiveValue = ""
	for _,value in pairs(area) do
		tosend =  tosend .. fromhex(value)
	end
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** ResetSensors = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** ResetSensors receive ERROR! ")
		return nil
	end
end

---------------------------------------------------------
---------------Panel Command Group----------------------
---------------------------------------------------------

local function PanelCapacities()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("1f")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** PanelCapacities = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** PanelCapacities receive ERROR! ")
		return nil
	end
end

local function RequestUser(userID) --  "01"
	local tosend = fromhex("40")
	tosend =  tosend .. fromhex(userID)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestUser = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestUser receive ERROR! ")
		return nil
	end
end

local function RequestAlarmMemoryPriorities()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("21")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestAlarmMemoryPriorities = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestAlarmMemoryPriorities receive ERROR! ")
		return nil
	end
end

local function RequestAlarmAreasByPriority(value)
	local tosend = ""
	local receiveValue = ""
	tosend = tosend .. fromhex("22") .. fromhex(value)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestAlarmAreasByPriority = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestAlarmAreasByPriority receive ERROR! ")
		return nil
	end
end

local function SilenceBells(bell) -- multible values allowed 2 bytes ex: "0008"
	local tosend = fromhex("19")
	local receiveValue = ""
	tosend =  tosend .. fromhex(bell)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** SilenceBells = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** SilenceBells receive ERROR! ")
		return nil
	end
end
local function trySilenceBells(bell)
	local responseString = ""
	local responseArray = {}
	responseString = SilenceBells(bell)
	if responseString then
		responseArray = parse(responseString,{})
		if responseArray[3] == "FC" then
			debug("(trySilenceBells) : Bell Silenced")
			displayMessage("Bell Silenced", TASK.ERROR)
		else
			debug("(trySilenceBells) : ERROR : Cannot Silence Bell")
			displayMessage(": ERROR : Cannot Silence Bell", TASK.ERROR)
		end
	end
end
---------------------------------------------------------
---------------Area Command Group----------------------
---------------------------------------------------------

local function RequestConfiguredAreas()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("24")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestConfiguredAreas = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestConfiguredAreas receive ERROR! ")
		return nil
	end
end

local areaStatusDisplay = {
	["00"] 				= "Unknown",
	["01"] 				= "Armed", --"All On",
	["02"] 				= "StayInstant", --"Part On instant",
	["03"] 				= "Stay", --"Part On Delay",
	["04"] 				= "Disarmed",
	["05"] 				= "Armed", --"All On Entry Delay",
	["06"] 				= "Stay", --"Part On Entry Delay",
	["07"] 				= "Armed", --"All On Exit Delay",
	["08"] 				= "Stay", --"Part On Exit Delay",
	["09"] 				= "ArmedInstant" --"All On Instant Armed"

}

local areaStatus = {
	["00"] 				= "Unknown",
	["01"] 				= "All On",
	["02"] 				= "Part On instant",
	["03"] 				= "Part On Delay",
	["04"] 				= "Disarmed",
	["05"] 				= "All On Entry Delay",
	["06"] 				= "Part On Entry Delay",
	["07"] 				= "All On Exit Delay",
	["08"] 				= "Part On Exit Delay",
	["09"] 				= "All On Instant Armed"

}
local function RequestAreaStatus(area) -- multible values allowed
	local tosend = fromhex("26")
	local receiveValue = ""
	for _,value in pairs(area) do
		tosend = tosend .. fromhex(value)
	end
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestAreaStatus = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestAreaStatus receive ERROR! ")
		return nil
	end
end

local armingTypeArray = {
	["Disarm"] 				= "01",
	["MasterInstantArm"] 	= "02", -- ArmedInstant
	["MasterDelayArm"] 		= "03", -- Armed
	["PerimeterInstantArm"] = "04", -- StayInstant
	["PerimeterDelayArm"] 	= "05", -- Stay
	["ForceArm"] 			= "06"  -- Force

}

local armingTypeArray2 = {
	["01"] = "Disarmed",
	["0A"] = "Arm Stay",
	["0B"] = "Arm Stay2",
	["0C"] = "Arm Away",
}

local function ArmPanelAreas(armingType, area)
	local tosend = ""
	local receiveValue = ""

	tosend = fromhex("27") .. fromhex(armingType) ..  fromhex(area)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** ArmPanelAreas = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** ArmPanelAreas receive ERROR! ")
		return nil
	end
end

local function RequestAreasNotReadyToArm()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("28")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestAreasNotReadyToArm = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestAreasNotReadyToArm receive ERROR! ")
		return nil
	end
end

local function RequestAreaText(area, language) -- area 2bytes; language 1byte 00 - English, 01 - Language 2

	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("29") .. fromhex(area) .. fromhex(language)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestAreaText = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestAreaText receive ERROR! ")
		return nil
	end
end

---------------------------------------------------------
---------------Door Command Group----------------------
---------------------------------------------------------

local function RequestConfiguredDoors()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("2B")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestConfiguredDoors = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestConfiguredDoors receive ERROR! ")
		return nil
	end
end

--[[
Door Status Value
Door Unlocked Bit 7 = true
Door Secured Bit 6 = true
Door in Learn Mode Bit 5 = true
Door in Diagnostic Mode Bit 4 = true
Door in SDI Failure Bit 2 = true
Not defined Bit 3, Bit 1 & Bit 0

]]

local function RequestDoorStatus(door) -- multible values allowed 2 bytes ex: {"01","03"}
	local tosend = fromhex("2C")
	local receiveValue = ""
	for _,value in pairs(door) do
		tosend =  tosend .. fromhex(value)
	end
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestDoorStatus = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestDoorStatus receive ERROR! ")
		return nil
	end
end

local doorState = {
	["NoAction"] 				= "00",
	["Cycle"] 					= "01",
	["UnlockDoor"] 				= "02",
	["TerminateUnlockMode"] 	= "03",
	["Secure"] 					= "04",
	["Terminate Secure Mode"] 	= "05"
}

local function SetDoorState(door) -- door 1byte "03"
 	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("2D") .. fromhex(door)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** SetDoorState = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** SetDoorState receive ERROR! ")
		return nil
	end
end

local function RequestDoorText(door, language) -- door 1bytes; language 1byte 00 - English, 01 - Language 2
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("2E") .. fromhex(door) .. fromhex(language)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestDoorText = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestDoorText receive ERROR! ")
		return nil
	end
end

---------------------------------------------------------
---------------Output Command Group----------------------
---------------------------------------------------------

local function RequestConfiguredOutputs()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("30")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestConfiguredOutputs = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestConfiguredOutputs receive ERROR! ")
		return nil
	end
end

local function RequestOutputStatus()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("31")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestOutputStatus = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestOutputStatus receive ERROR! ")
		return nil
	end
end

local function SetOutputState(output, state) -- output 1byte "03", status: "00" -> "Inactive" "01" -> "Active"
 	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("32") .. fromhex(output) .. fromhex(state)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** SetOutputState = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** SetOutputState receive ERROR! ")
		return nil
	end
end

local function RequestOutputText(output, language) -- output 1bytes; language 1byte 00 - English, 01 - Language 2
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("33") .. fromhex(output) .. fromhex(language)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestOutputText = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestOutputText receive ERROR! ")
		return nil
	end
end

---------------------------------------------------------
---------------Point Command Group----------------------
---------------------------------------------------------

local function RequestConfiguredPoints()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("35")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestConfiguredPoints = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestConfiguredPoints receive ERROR! ")
		return nil
	end
end

local function RequestPointsInArea(area) -- area 2bytes
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("36") .. fromhex(area)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestPointsInArea = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestPointsInArea receive ERROR! ")
		return nil
	end
end

local function RequestFaultedPoints()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("37")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestFaultedPoints = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestFaultedPoints receive ERROR! ")
		return nil
	end
end

local pointStatus = {
	["00"] 				= "Unassigned",
	["01"] 				= "Short",
	["02"] 				= "Open",
	["03"] 				= "Normal",
	["04"] 				= "Missing",
}

local function RequestPointsStatus(points) -- multible values allowed 2 bytes ex: "0008"
	local tosend = fromhex("38")
	local receiveValue = ""
	for _,value in pairs(points) do
		tosend =  tosend .. fromhex(value)
	end
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestPointsStatus = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestPointsStatus receive ERROR! ")
		return nil
	end
end

local function RequestBypassedPoints()
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("39")
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestBypassedPoints = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestBypassedPoints receive ERROR! ")
		return nil
	end
end

local function BypassPoints(points) -- multible values allowed 2 bytes ex: "0008"
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("3A")
	for _,value in pairs(points) do
		tosend = tosend .. fromhex(value)
	end
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** BypassPoints = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** BypassPoints receive ERROR! ")
		return nil
	end
end

local function UnbypassPoints(points) -- multible values allowed 2 bytes ex: "0008"
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("3B")
	for _,value in pairs(points) do
		tosend = tosend .. fromhex(value)
	end
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** UnbypassPoints = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** UnbypassPoints receive ERROR! ")
		return nil
	end
end

local function RequestPointText(point, language) -- point 2bytes; language 1byte 00 - English, 01 - Language 2
	local tosend = ""
	local receiveValue = ""
	tosend = fromhex("3C") .. fromhex(point) .. fromhex(language)
	send(tosend,peer)
	receiveValue = receive()
	if receiveValue then
		debug ("*** RequestPointText = " .. tohex(receiveValue))
		return tohex(receiveValue)
	else
		debug ("*** RequestPointText receive ERROR! ")
		return nil
	end
end

---------------------------------------------------------
---------------------Getting Infos-----------------------
---------------------------------------------------------
local function getCapacities()
	local responseString
	local responseArray = {}

	responseString = PanelCapacities()
	if responseString then
		responseArray = parse(responseString,{7,22})
		numberOfAreas = tonumber(responseArray[4],16)
		numberOfPoints = tonumber(responseArray[5],16)
		numberOfOutputs = tonumber(responseArray[6],16)
		numberOfUsers = tonumber(responseArray[7],16)
		numberOfKeypads = tonumber(responseArray[8],16)
		numberOfDoors = tonumber(responseArray[9],16)
		numberOfEventRecordSize = tonumber(responseArray[10],16)
	else
		debug("(getCapacities) : ERROR :  UNable to get Panel Capacities!")
	end
end

local function getUsers()
	local responseString
	local responseArray = {}
	local i
	local count = 1
	g_users = {}
	for i=1,numberOfUsers do
		responseString = RequestUser(string.gsub(string.format("%04s",tostring(ToHexNumber(i))),"%s","0"))
		if responseString then
			responseArray = parse(responseString,{7,10,13,20})
			if responseArray[3] == "FE" then
				g_users[count] = {}
				g_users[count].ID = responseArray[4]
				g_users[count].passcode = responseArray[6] .. responseArray[7]
				g_users[count].passcode = string.gsub(g_users[count].passcode,"F", "")
				g_users[count].name = ""
				for j=8,#responseArray do
					if responseArray[j] ~= "00" then
						g_users[count].name = g_users[count].name .. string.char(tonumber(responseArray[j],16))
					else
						break
					end
				end
				count = count + 1
			end
		end
	end
end

local function getAreas()
	local responseString
	local responseArray = {}
	local array = {}
	local i = 0
	local count = 1
	local value = ""
	local areaCount = 1
	responseString = RequestConfiguredAreas()
	if responseString then
		responseArray = parse(responseString,{})
		for i = 4,#responseArray do
			value = Hex2Bin(responseArray[i])
			for word in string.gmatch(value, ".") do
				array[count] = word
				count = count + 1
			end
			for key,val in pairs(array) do

				if val == "1" then
					g_areas[areaCount] = {}
					g_areas[areaCount].panelID = key
					areaCount = areaCount + 1
				end
			end
		end
	else
		debug("(getAreas) : ERROR :  Unable to get Areas!")
	end

	for key,val in pairs(g_areas) do
		local name = ""
		responseString = RequestAreaText(string.gsub(string.format("%04s",val.panelID),"%s","0"),"00")
		if responseString then
			responseArray = parse(responseString,{})
			for i = 4, #responseArray do
				if i ~= #responseArray then
					name = name .. string.char(tonumber(responseArray[i],16))
				end
			end
			g_areas[key].name = name
		else
			debug("(getAreas) : ERROR :  Unable to get Area text!")
		end
	end
end

local function getPoints()
	local responseString
	local responseArray = {}
	local array = {}
	local i = 0
	local count = 1
	local value = ""
	local pointCount = 1
	responseString = RequestConfiguredPoints()
	if responseString then
		responseArray = parse(responseString,{})
		for i = 4,#responseArray do
			value = Hex2Bin(responseArray[i])
			for word in string.gmatch(value, ".") do
				array[count] = word
				count = count + 1
			end
			for key,val in pairs(array) do
				if val == "1" then
					g_points[pointCount] = {}
					g_points[pointCount].panelID = key
					pointCount = pointCount + 1
				end
			end
		end
	else
		debug("(getPoints) : ERROR :  Unable to get Configured Points!")
	end
	for key,val in pairs(g_points) do
		local name = ""
		responseString = RequestPointText(string.gsub(string.format("%04s",tostring(ToHexNumber(tonumber(val.panelID)))),"%s","0"),"00")
		if responseString then
			responseArray = parse(responseString,{})
			for i = 4, #responseArray do
				if i ~= #responseArray then
					name = name .. string.char(tonumber(responseArray[i],16))
				end
			end
			g_points[key].name = name
		else
			debug("(getPoints) : ERROR :  Unable to get Points Text!")
		end
	end
end

local function getDoors()
	local responseString
	local responseArray = {}
	local array = {}
	local i = 0
	local count = 1
	local value = ""
	responseString = RequestConfiguredDoors()
	if responseString then
		responseArray = parse(responseString,{})
		for i = 4,#responseArray do
			value = Hex2Bin(responseArray[i])
			for word in string.gmatch(value, ".") do
				array[count] = word
				count = count + 1
			end
			for key,val in pairs(array) do
				if val == "1" then
					table.insert(g_doors, key)
				end
			end
		end
	else
		debug("(getDoors) :  ERROR : Unable to get Configured Doors!")
	end
end

local function getOutputs()
	local responseString
	local responseArray = {}
	local array = {}
	local i = 0
	local count = 1
	local value = ""
	local outputsCount = 1
	responseString = RequestConfiguredOutputs()
	if responseString then
		responseArray = parse(responseString,{})
		for i = 4,#responseArray do
			value = Hex2Bin(responseArray[i])
			for word in string.gmatch(value, ".") do
				array[count] = word
				count = count + 1
			end
		end
		for key,val in pairs(array) do
				if val == "1" then
					g_outputs[outputsCount] = {}
					g_outputs[outputsCount].panelID = key
					g_outputs[outputsCount].status = "0"
					g_outputs[outputsCount].name = string.format("Solution Output %d", key)
					outputsCount = outputsCount + 1
				end
		end
	else
		debug("(getOutputs): ERROR : Unable to get Configured Points!")
	end
end

local function getAreaState(id)
	local area = ""
	local responseString
	local responseArray = {}
	local areaArray = {}

	area = g_areas[id].panelID
	area = string.gsub(string.format("%04s",area),"%s","0")
	table.insert(areaArray,area)
	responseString = RequestAreaStatus(areaArray)
	if responseString then
		responseArray = parse(responseString,{7,10})
		return areaStatus[responseArray[5]], areaStatusDisplay[responseArray[5]]
	else
		return nil, nil
	end
end

local function getPointState(id)
	local point = id
	local responseString
	local responseArray = {}
	local pointArray = {}

	point = string.gsub(string.format("%04s",tostring(ToHexNumber(tonumber(point)))),"%s","0")
	table.insert(pointArray,point)
	responseString = RequestPointsStatus(pointArray)
	if responseString then
		responseArray = parse(responseString,{7,10})
		return pointStatus[responseArray[5]]
	else
		debug("(getPointState) : ERROR : Unable to get Point Status!")
		return nil
	end
end

local function getDoorState(id)
	local door = id
	local responseString
	local responseArray = {}
	local pointArray = {}

	door = string.gsub(string.format("%02s",tostring(ToHexNumber(tonumber(door)))),"%s","0")
	table.insert(pointArray,door)
	responseString = RequestDoorStatus(pointArray)
	if responseString then
		responseArray = parse(responseString,{})
		return pointStatus[responseArray[5]]
	else
		debug("(getDoorState) : ERROR : Unable to get Door Status!")
		return nil
	end
end

local function getOutputState()
	local responseString
	local responseArray = {}
	local array = {}
	local count = 1

	responseString = RequestOutputStatus()
	if responseString then
		responseArray = parse(responseString,{})
		if #responseArray > 3 then
			for i = 4,#responseArray do
				value = Hex2Bin(responseArray[i])
				for word in string.gmatch(value, ".") do
					array[count] = word
					count = count + 1
				end
			end
		else
			for key,val in pairs(g_outputs) do
				g_outputs[key].status = "0"
			end
		end

		for key,val in pairs(g_outputs) do
			for k,v in pairs(array) do
				if val.panelID == k then
					g_outputs[key].status = v
				end
			end
			
			local outputState = luup.variable_get(SID.SWP_SID, "Status", val.devId) or ""
			
			if outputState ~= "" then
				if outputState ~= tostring(val.status) then
					luup.variable_set(SID.SWP_SID, "Status", val.status, val.devId)
					luup.variable_set(SID.SWP_SID, "Target", val.status, val.devId)
					debug("(getOutputState) : Output state for " .. val.panelID .. " has been set to " .. val.status)
				else
					debug("(getOutputState) : No need to update the state for Output " .. val.panelID)
				end
			else
				luup.variable_set(SID.SWP_SID, "Status", val.status, val.devId)
				luup.variable_set(SID.SWP_SID, "Target", val.status, val.devId)
			end
		end
	else
		debug("(getOutputState) : ERROR : Unable to get Outputs Status!")
	end
end

local function getHistoryRaw(noEvents, startEvent)
	local responseString = ""
	local responseArray = {}

	responseString = RequestRawHistoryEvents(noEvents,startEvent)
	if responseString then
		responseArray = parse(responseString,{})
		for k,v in pairs(responseArray) do
			log(v)
		end
	end
end

local function getHistoryText(noEvents, startEvent)
	local responseString = ""
	local responseArray = {}

	responseString = RequestTextHistoryEvents(noEvents,startEvent)
	if responseString then
		responseArray = parse(responseString,{})
		for k,v in pairs(responseArray) do
			log(v)
		end
	end
end

local function getHistoryIndexLimits()
	local responseString = ""
	local responseArray = {}
	local value = ""
	responseString = RequestTextHistoryEvents("01","00000000")
	if responseString then
		responseArray = parse(responseString,{})
		for key,val in pairs(responseArray) do
			if key > 4 and key <= 8 then
				value = value .. val
			end
		end
	end
	historyMin = tonumber(value,16)
	value = ""
	responseString = RequestTextHistoryEvents("01","FFFFFFFF")
	if responseString then
		responseArray = parse(responseString,{})
		for key,val in pairs(responseArray) do
			if key > 4 and key <= 8 then
				value = value .. val
			end
		end
	end
	historyMax = tonumber(value,16)
end

function sendBypassPoints(points)

	local array = {}
	local count  = 1
	local responseString
	local responseArray = {}

	debug("(sendBypassPoints) : Point/s requested for bypass : " .. points)

	for word in string.gmatch(points, "(%d+)") do
			array[count] = word
			count = count + 1
	end
	for k,v in pairs(array) do
		array[k] = string.gsub(string.format("%04s",tostring(ToHexNumber(tonumber(v)))),"%s","0")
	end
	responseString = BypassPoints(array)
	if responseString then
		responseArray = parse(responseString,{})
	else
		debug("(sendBypassPoints) : ERROR : Unable to Bypass Point/s !")
	end
end

function sendUnbypassPoints(points)
	local array = {}
	local count  = 1
	local responseString
	local responseArray = {}

	debug("(sendUnbypassPoints) : Point/s requested for unbypass : " .. points)

	for word in string.gmatch(points, "(%d+)") do
			array[count] = word
			count = count + 1
	end
	for k,v in pairs(array) do
		array[k] = string.gsub(string.format("%04s",tostring(ToHexNumber(tonumber(v)))),"%s","0")
	end
	responseString = UnbypassPoints(array)
	if responseString then
		responseArray = parse(responseString,{})
	else
		debug("(sendUnbypassPoints) : ERROR :  Unable to Unbypass Point/s!")
	end
end

local alarmPriority = {
	["0001"] = "01",
	["0002"] = "02",
	["0004"] = "03",
	["0008"] = "04",
	["0010"] = "05",
	["0020"] = "06",
	["0040"] = "07",
	["0080"] = "08",
	["0100"] = "09",
	["0200"] = "10",
}

local alarmPriorityText = {
	["0001"] = "Burglary Trouble",
	["0002"] = "Burglary Supervisory",
	["0004"] = "Gas Trouble",
	["0008"] = "Gas Supervisory",
	["0010"] = "Fire Trouble",
	["0020"] = "Fire Supervisory",
	["0040"] = "Burglary Alarm",
	["0080"] = "Personal Emergency",
	["0100"] = "Gas Alarm",
	["0200"] = "Fire Alarm",
}

function getAlarmReport()
	local responseString
	local responseArray = {}
	local alarmType = ""
	local value
	local array = {}
	local count = 1
	responseString = RequestAlarmMemoryPriorities()
	if responseString then
		responseArray = parse(responseString,{7,10})
		if responseArray[4] == "0000" then
			for key,val in pairs(g_areas) do
				local alarm = ""
				alarm = luup.variable_get(SID.PARTITION,"Alarm", g_areas[key].devId) or ""
				if alarm == "" then
					luup.variable_set(SID.PARTITION,"Alarm","None", g_areas[key].devId)
				else
					if alarm == "Active" then
						luup.variable_set(SID.PARTITION,"Alarm","None", g_areas[key].devId)
					end
				end
			end
			debug("(getAlarmReport) :  We don't have alarms in the system!")
		else
			alarmType = responseArray[4]
			responseString = ""
			responseArray = {}
			responseString = RequestAlarmAreasByPriority(alarmPriority[alarmType])
			if responseString then
				responseArray = parse(responseString,{})
				for i = 4,#responseArray do
					value = Hex2Bin(responseArray[i])
					for word in string.gmatch(value, ".") do
						array[count] = word
						count = count + 1
					end
				end

				for key,val in pairs(array) do
					local alarm = ""
					if val == "1" then
						alarm = luup.variable_get(SID.PARTITION,"Alarm", g_areas[key].devId) or ""  
						if type(g_areas[key]) == "table" then
							if alarm == "" then
								luup.variable_set(SID.PARTITION,"Alarm","Active", g_areas[key].devId)
							else
								if alarm == "None" then
									luup.variable_set(SID.PARTITION,"Alarm","Active", g_areas[key].devId)
								end
							end
							flagAlarm = true
							log(alarmPriorityText[alarmType] .. " for area : " .. g_areas[key].name)
							displayMessage(alarmPriorityText[alarmType] .. " for area : " .. g_areas[key].name, TASK.ERROR)
							-- add alert for alarms
							if not flagAlert then
								local url = "http://127.0.0.1:3480/data_request?id=add_alert&device="..g_areas[key].devId.."&type=12&source=3&severity=0&code="..UrlEncode("Remote Security Plus Alarm").. "&description="..UrlEncode("We have a "..alarmPriorityText[alarmType].." for area "..g_areas[key].name)
								--local url = "http://"..unitIp..":3480/data_request?id=add_alert&device="..lug_device.."&type=12&source=3&severity=0&code=RSP_Alarm&description=Alarm"
								local status, content = luup.inet.wget(url,15)
								if not content then
									debug("(getAlarmReport) : ERROR :  Alert could NOT be set!")
								else
									debug("(getAlarmReport) : OK :  Alert has been set!")
									flagAlert = true
								end
							end
						end
					else
						if g_areas[key] then
							luup.variable_set(SID.PARTITION,"Alarm","None", g_areas[key].devId)
						end
					end
				end
			else
				debug("(getAlarmReport) : ERROR :  Unable to get Alarm Areas by Priority!")
			end
		end
	else
		debug("(getAlarmReport) : ERROR :  Unable to get Alarm Memory Proprieties!")
	end
end
---------------------------------------------------------
---------------Action Implementations--------------------
---------------------------------------------------------
function clearTask()
	if (os.time() - g_lastTask >= DISPLAY_SECONDS) then
		luup.task("Clearing...", TASK.SUCCESS, "BoschPlugin", g_taskHandle)
	end
	debug("(clearTask) : Clearing task... ")
end

function requestQuickArmMode (device, state)
	debug("(requestQuickArmMode) : Quick Arm not supported.")
	displayMessage("Quick Arm not supported.", TASK.ERROR)
	return false
end

function requestPanicMode (device, state)
	debug("(requestPanicMode) : Panic Modes not supported.")
	displayMessage("Panic Modes not supported.", TASK.ERROR)
	return false
end

function setTarget (newTargetValue, device)
	local output = ""
	local state = ""
	local response = ""
	local key
	for k,v in pairs(g_outputs) do
		if device == v.devId then
			key = k
			output = string.gsub(string.format("%02s",tostring(ToHexNumber(v.panelID))),"%s","0")
		end
	end
	state = string.gsub(string.format("%02s",tostring(newTargetValue)),"%s","0")
	response = SetOutputState(output, state)
	if response then
		g_outputs[key].status = tostring(newTargetValue)
		luup.variable_set(SID.SWP_SID, "Status", newTargetValue, device)
		luup.variable_set(SID.SWP_SID, "Target", newTargetValue, device)
	else
		debug("(setTarget) : Cannot update Output status, Please retry!")
		displayMessage("Cannot update Output status, Please retry!", TASK.ERROR)
	end
end

local function setAreaState (partNo, variable, value, valueDisplay)
	debug("(setAreaState) : Set variable '".. variable .."' = '".. valueDisplay .."' aka : '" .. value .. "' for partition ".. partNo .." (device ".. tostring(g_areas[partNo].devId) ..").")
	luup.variable_set(SID.PARTITION, variable, value, g_areas[partNo].devId)
	if value == "Disarmed" then
		luup.variable_set(SID.PARTITION, "ArmMode", "Disarmed", g_areas[partNo].devId)
	else
		luup.variable_set(SID.PARTITION, "ArmMode", "Armed", g_areas[partNo].devId)
	end
end

function requestArmMode (device, state, pinCode)
	local partitionNo = 0
	local partitionPannelNo = 0
	for key,val in pairs(g_areas) do
		if device == val.devId then
			partitionNo = key
			partitionPannelNo = tonumber(val.panelID,10)
		end
	end
	if (partitionNo == 0) then
		debug("(requestArmMode) : ERROR : Invalid partition number. The altid of device ".. device .." is '".. luup.devices[device].id .."'.")
		displayMessage("Error storing the PIN code.", TASK.ERROR)
		return
	end

	debug("(requestArmMode) : Request to arm partition ".. g_areas[partitionNo].name .." (device ".. device ..") to state '".. state .."'.")
	--[[
	pinCode = tonumber(pinCode) or g_pinCodes[tonumber(g_areas[partitionNo].panelID)]
	if (not pinCode) then
		debug("(requestArmMode) : ERROR : PIN code required.")
		displayMessage("PIN code required.", TASK.ERROR)
		return false
	end
	-- ]]
	local arming
	if (state == "Disarmed") then
		arming = "01" -- Disarm
	elseif (state == "Armed") then
		arming = "0C" -- Away
	elseif (state == "Stay") then
		arming = "0A" -- 
	elseif (state == "Stay2") then
		arming = "0B" -- 
	else
		debug("(requestArmMode) : ERROR : Invalid state requested.")
		displayMessage("Invalid state requested.", TASK.ERROR)
		return false
	end

	local responseString
	local responseArray = {}
	local mask = "" -- "80"
	for i = 1,8 do
		if i == partitionPannelNo then
			mask = mask .. "1"
		else
			mask = mask .. "0"
		end
	end
	mask = string.format("%02s",Bin2Hex(mask))
	responseString = ArmPanelAreas(arming, mask)
	if responseString then
		responseArray = parse(responseString,{})
		if responseArray[3] == "FC" then
			local detailedArmMode = luup.variable_get(SID.PARTITION, "DetailedArmMode", g_areas[partitionNo].devId)
			if detailedArmMode ~= state then
				luup.variable_set(SID.PARTITION, "DetailedArmMode", state, g_areas[partitionNo].devId)
			end
			if arming == "01" then
				local armMode = luup.variable_get(SID.PARTITION, "ArmMode", g_areas[partitionNo].devId)
				if armMode ~= "Disarmed" then
					luup.variable_set(SID.PARTITION, "ArmMode", "Disarmed", g_areas[partitionNo].devId)
				end
				if flagAlarm then
					trySilenceBells(string.gsub(string.format("%04s",tostring(g_areas[partitionNo].panelID)),"%s","0"))
					flagAlarm = false
					flagAlert = false
				end
			else
				local armMode = luup.variable_get(SID.PARTITION, "ArmMode", g_areas[partitionNo].devId)
				if armMode ~= "Armed" then
					luup.variable_set(SID.PARTITION, "ArmMode", "Armed", g_areas[partitionNo].devId)
				end
			end
			debug("(requestArmMode) : Successfully changed arming state to " ..  armingTypeArray2[arming])
			displayMessage("Successfully changed arming state to " ..  armingTypeArray2[arming], TASK.ERROR)
		else
			if responseArray[3] == "FD" and responseArray[4] == "08" then
				debug("(requestArmMode) : ERROR :  Cannot arm the partition, one or more points are FAULTED!")
				displayMessage("Cannot arm the partition, one or more points are FAULTED!", TASK.ERROR)
			else
				debug("(requestArmMode) : ERROR : we should not be here!")
			end
		end
	else
		debug("(requestArmMode) : ERROR : Unable to Arm Panel Areas!")
	end
end

function storePinCode (device, pinCode)
	local partitionNo = 0
	local flag  = false
	for key,val in pairs(g_areas) do
		if device == val.devId then
			partitionNo = key
		end
	end

	pinCode = pinCode or ""
	if (pinCode == "") then
		debug("(storePinCode) : ERROR : Empty PIN code field.")
		displayMessage("Empty PIN code field.", TASK.ERROR)
		return
	end

	if (pinCode:match("^%d%d%d%d%d%d$") == nil) then
		debug("(storePinCode) : ERROR : Insert a 6 digits pincode!")
		displayMessage("Invalid PIN code.", TASK.ERROR)
		return
	end
	for key,val in pairs(g_users) do
		if pinCode == val.passcode then
			flag = true
		end
	end
	if flag then
		g_pinCodes[tonumber(g_areas[partitionNo].panelID)] = pinCode
		debug("(storePinCode) : PIN code stored for partition " .. g_areas[partitionNo].name .. ".")
	else
		debug("(storePinCode) : ERROR : Invalid PIN code.")
	end
end


function clearPinCode (device)
	local partitionNo = 0
	for key,val in pairs(g_areas) do
		if device == val.devId then
			partitionNo = key
		end
	end

	if (partitionNo == 0) then
		debug("(clearPinCode) : ERROR : Invalid partition number. The altid of device ".. device .." is '".. luup.devices[device].id .."'.")
		displayMessage("Error storing the PIN code.", TASK.ERROR) -- assert level debug message? O.o
		return
	end

	g_pinCodes[tonumber(g_areas[partitionNo].panelID)] = nil
	debug("(clearPinCode) : PIN code cleared for partition " .. g_areas[partitionNo].name .. ".")

end

function setHistoryInterval()
	luup.variable_set("urn:micasaverde-com:serviceId:BoschSolutionAlarmPanel1","Interval", tostring(historyMin) .. " - " .. tostring(historyMax), lul_device)
end

function getHistory(noEvents, startEvent)
	local responseString = ""
	local responseArray = {}
	historyLog = "<table border=\"1\" cellpadding=\"8\"><tr><td>EVENTS</td></tr>"
	noEvents = tonumber(noEvents)
	startEvent = tonumber(startEvent)

	-- [[
	if noEvents < 1 or noEvents >  math.abs(historyMax - historyMin) then
		local msg = "Number of events must be at least 1 and not grater then " .. math.abs(historyMax - historyMin)
		log(msg)
		displayMessage(msg, TASK.ERROR)
		return
	end

	if startEvent < historyMin or startEvent > historyMax then
		local msg = "Start Event  must be grater then " .. historyMin  .. " and smaller then " .. historyMax
		log(msg)
		displayMessage(msg, TASK.ERROR)
		return
	end
	-- ]]
	for i = startEvent,startEvent + noEvents - 1 do
		responseString = RequestTextHistoryEvents("01",string.gsub(string.format("%08s",tostring(ToHexNumber(i))),"%s","0"))
		if responseString then
			local message = ""
			responseArray = parse(responseString,{})
			for key,val in pairs(responseArray) do
				if key >=  9 then
					if tonumber(val,16) ~= 0 then
						message = message .. string.char(tonumber(val,16))
					end
				end
			end
			debug("*******History[" .. i .. "] = [" ..  message .. "] " )

			historyLog = historyLog .. "<tr><td>" .. message .. "</td></tr>"
		end
	end
	historyLog = historyLog .. "</table>"
	luup.variable_set(SID.BOSCH,"HistoryLog",historyLog,lul_device)
end
---------------------------------------------------------
---------------Initialization Functions------------------
---------------------------------------------------------
local function getChildDevices(device)
	for dev, attr in pairs(luup.devices) do
		if (attr.device_num_parent == device) then
			local zoneNo = attr.id:match("^bosch_point_(%d+)")
			local partitionNo = attr.id:match("^bosch_area_(%d+)")
			local outputNo = attr.id:match("^bosch_output_(%d+)_")
			for k,v in pairs(g_areas) do
				if partitionNo == tostring(v.panelID) then
					partitionNo = tostring(k)
				end
			end
			if (zoneNo ~= nil) then
				zoneNo = tonumber(zoneNo, 10)
				g_points[zoneNo].devId = dev
			elseif (partitionNo ~= nil) then
					partitionNo = tonumber(partitionNo, 10)
					g_areas[partitionNo].devId = dev
			else
				if (outputNo ~= nil) then
					outputNo = tonumber(outputNo, 10)
					for key,val in pairs(g_outputs) do
						if outputNo == tonumber(val.panelID) then
							g_outputs[key].devId = dev
						end
					end
				end
			end
		end
	end
end

local function appendPoints()
	local count = 0
	for i, v in ipairs(g_points) do
		debug("(appendPoints) : Appending zone ".. i ..".")
		local device = getDeviceType(i)
		luup.chdev.append(lug_device, g_appendPtr, "bosch_point_"..i, "Point ".. i ..": ".. v.label, device.type, device.file, nil, nil, false)
		if count > 20 then
			debug("(appendPoints) : Possible error in generating points, more then 20 devices were generated!!!")
			FLAGS.ZONES = true
			return
		end
	end
	FLAGS.ZONES = true
end

local function appendAreas()
	local count = 0
	for i, v in ipairs(g_areas) do
		debug("(appendAreas) : Appending partition ".. i ..".")
		luup.chdev.append(lug_device, g_appendPtr, "bosch_area_".. v.panelID, "Area ".. v.panelID ..": ".. v.name, "urn:schemas-micasaverde-com:device:BoschSolutionAlarmPartition:1", "D_BoschSolutionAlarmPartition1.xml", nil, nil, false)
		count = count + 1
		if count > 20 then
			debug("(appendAreas) : Possible error in generating areas, more then 20 devices were generated!!!")
			FLAGS.PARTITIONS = true
			return
		end
	end
	FLAGS.PARTITIONS = true
end

local function appendOutputs()
	local count  = 0
	for i, v in ipairs(g_outputs) do
		debug("(appendOutputs) : Appending output ".. v.name ..".")
		luup.chdev.append(lug_device, g_appendPtr, "bosch_output_".. v.panelID .. "_" .. v.name, v.name, DEVICE_TYPES.BINARY_LIGHT, DEVICE_FILES.BINARY_LIGHT, nil, nil, false)
		count = count + 1
		if count > 20 then
			debug("(appendOutputs) : Possible error in generating outputs, more then 20 devices were generated!!!")
			return
		end
	end
end

local function getInfos(device)	--[[ DebugMode enabled until the release  192
	local debugMode = luup.variable_get(SID.BOSCH, "DebugMode", device) or ""
	if (debugMode ~= "") then
		DEBUG_MODE = (debugMode == "1") and true or false
	else
		luup.variable_set(SID.BOSCH, "DebugMode", (DEBUG_MODE and "1" or "0"), device)
	end
	--]]
	DEBUG_MODE = true
	debug("(getDebugMode) : Debug mode "..(DEBUG_MODE and "enabled" or "disabled")..".")

	local polling_rate = luup.variable_get(SID.BOSCH, "POLLING_RATE", device) or ""
	if (polling_rate ~= "") then
		POLLING_RATE = tonumber(polling_rate)
	else
		luup.variable_set(SID.BOSCH, "POLLING_RATE", POLLING_RATE, device)
	end
	debug("(getDebugMode) : POLLING_RATE = " .. POLLING_RATE )
	
	local safestorage = luup.variable_get(SID.BOSCH,"SafeStorage",lug_device) or ""
	if safestorage == "" then
		luup.variable_set(SID.BOSCH,"SafeStorage","0",lug_device)
	end
	
	local passcode = luup.variable_get(SID.BOSCH,"Passcode",lug_device) or ""
	if passcode == "" then 
		luup.variable_set(SID.BOSCH,"Passcode","0000000000",lug_device)
	end

	local pincode = luup.variable_get(SID.BOSCH,"Pincode",lug_device) or ""
	if pincode == "" then 
		luup.variable_set(SID.BOSCH,"Pincode","2580",lug_device)
	end
	
	local trash
	ipAddress, trash, ipPort = string.match(luup.devices[device].ip, "^([%w%.%-]+)(:?(%d-))$")
	if ipAddress and ipAddress ~= "" then
		if ipPort==nil or ipPort == "" then
			ipPort = "7700"
		end
		return true
	else
		debug("(getDebugMode) : ERROR : Insert IP address!")
		return false
	end
end

function updateState()
	debug("(updateState) : OK : Start updating status ...")
	local StatusArea = ""
	local StatusAreaDisplay = ""
	for key,val in pairs (g_areas) do
		StatusAreaDisplay, StatusArea  = getAreaState(key)
		if StatusArea and StatusAreaDisplay then
			local detailedArmMode = luup.variable_get(SID.PARTITION, "DetailedArmMode", g_areas[key].devId)
			if detailedArmMode ~= StatusArea then
				setAreaState(key,"DetailedArmMode",StatusArea, StatusAreaDisplay)
			end
		else
			--debug("(updateState) : ERROR :  Cannot update state! Retrying !")
			debug("(updateState) : ERROR : Connection is lost! Trying to reconnect ...")
			displayMessage("Connection is lost! Trying to reconnect ...", TASK.ERROR)
			reconnect()
		end
	end
	getAlarmReport()
	getOutputState()
	luup.call_delay( "updateState", POLLING_RATE)
end

local function connectToPanel()
	local retry
	local i
	local flagConnect = false
	local flagPass = false
	local passcode
	
	local safestorage = luup.variable_get(SID.BOSCH,"SafeStorage",lug_device) or ""
	if safestorage == "" then
		luup.variable_set(SID.BOSCH,"SafeStorage","0",lug_device)
		safestorage = "0"
	end

	local pincode = luup.variable_get(SID.BOSCH,"Pincode",lug_device) or ""
	if pincode == "" then 
		luup.variable_set(SID.BOSCH,"Pincode","2580",lug_device)
	end

	if safestorage == "0" then
		passcode = luup.variable_get(SID.BOSCH,"Passcode",lug_device) or ""
		if passcode == "" then
			luup.variable_set(SID.BOSCH,"Passcode","0000000000",lug_device)
			debug("(connectToPanel) : ERROR : Password field have a default value!")
			displayMessage("Password field have a default value!", TASK.ERROR)
		else
			g_passcode = passcode
			flagPass = true
		end
	elseif safestorage == "1" then
		local file = io.open("/tmp/credential.txt")
		if file then
			for line in file:lines() do
				passcode = line:match('^(.-)$')
			end
			file:close()
			if passcode then
				g_passcode = passcode
				flagPass = true
			else
				debug("(connectToPanel) : ERROR : Please insert your passcode in file!")
				displayMessage("Please insert your passcode in file!", TASK.ERROR)
			end
		else
			debug("(connectToPanel) : ERROR : Couldn't open passcode file!")
			displayMessage("Couldn't open passcode file!", TASK.ERROR)
		end
	else
		debug("(connectToPanel) : ERROR : Invalid SafeStorage value. Please insert a valid one(0 or 1)!")
		displayMessage("Invalid SafeStorage value. Please insert a valid one(0 or 1)!", TASK.ERROR)
	end

	if flagPass then
		for i = 1, 10 do
			debug("(connectToPanel) : ERROR : Trying to reconnect[" .. i .. "]")
			retry =  peer:connect(ipAddress, ipPort)
			if retry then
				flagConnect = true
				break
			else
				debug("(connectToPanel) : ERROR : Cannot connect to panel!")
				g_taskHandle = luup.task("Cannot connect to panel!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
			end
		end
		if flagConnect then
			peer = assert( ssl.wrap(peer, params) )
			peer:settimeout(0.3)


			local succ, err = peer:dohandshake()
			while not succ do
				wait(peer, err)
				succ, err = peer:dohandshake()
			end
			debug("*** Handshake done ***")
			local responseArray = {}
			local responseString
			responseString = WhatAreYou()
			if responseString then
				responseArray = parse(responseString,{})
			else
				debug("(connectToPanel) : ERROR : Handshake ERROR!")
				g_taskHandle = luup.task("Handshake ERROR!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
			end
			responseString = PasscodeAndUserTypeCheck(g_passcode)
			if responseString then
				responseArray = parse(responseString,{})
				if responseArray[3] == "FE" then
					if responseArray[4] == "00" then
						debug("(connectToPanel) : ERROR :  Passcode and User Type Check ERROR! Another Session is opened, cannot authentify !")
						g_taskHandle = luup.task("Passcode and User Type Check ERROR! Another Session is opened, cannot authentify !", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
						flagPasscode = false
					elseif responseArray[4] == "01" then
						debug("(connectToPanel) : OK :  Passcode and User Type Check OK! continuing ...")
						
						responseString = PincodeCheck(pincode)
						if responseString then
							responseArray = parse(responseString,{})
							if responseArray[3] == "FE" then
								debug("(connectToPanel) : OK :  Pincode Check OK! continuing ...")
								flagPasscode = true
							else
								debug("(connectToPanel) : ERROR :  Pincode Check ERROR! response= " .. responseString)
								g_taskHandle = luup.task("Pincode Check ERROR!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
								flagPasscode = false
							end
						else
							debug("(connectToPanel) : ERROR :  Pincode Check ERROR!")
							g_taskHandle = luup.task("Pincode Check ERROR!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
							flagPasscode = false
						end
					else
						debug("(connectToPanel) : ERROR :  Unknown Passcode and User Type Check ERROR! FE : [" .. responseArray[4] .. "]")
						g_taskHandle = luup.task("Unknown Passcode and User Type Check ERROR!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
						flagPasscode = false
					end
				elseif responseArray[3] == "FD" then
					debug("(connectToPanel) : ERROR :  NAK received with error code " .. responseArray[4])
					flagPasscode = false
				elseif responseArray[3] == "FC" then
					debug("(connectToPanel) : ERROR :  ACK received")
					flagPasscode = false
				else
					debug("(connectToPanel) : ERROR :  Unknown ERROR (we should not be here)")
					flagPasscode = false
				end
			else
				debug("(connectToPanel) ERROR : Empty response from PasscodeAndUserTypeCheck!")
				g_taskHandle = luup.task("Unknown Passcode and User Type Check ERROR!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
			end
		else
			debug("(connectToPanel) : ERROR : Please check IP and PORT settings!")
			displayMessage("Please check IP and PORT settings!", TASK.ERROR_STOP)
		end
	else
		debug("(connectToPanel) : ERROR : Please check Passcode settings!")
		displayMessage("Please check Passcode settings!", TASK.ERROR)
	end
end

function reconnect()
	debug("(reconnect) : OK : Reconnecting ...")
	peer:close()
	peer = socket.tcp()
	connectToPanel()
end

function generateData()
	connectToPanel()
	if flagPasscode then
		getCapacities()
		--getUsers()
		getAreas()
		getPoints()
		getDoors()
		getOutputs()
	end
end

function Init(lul_device)
	local StatusArea = ""
	local StatusPoint = ""
	local errorParams
	lug_device = lul_device
	g_taskHandle = luup.task("Starting up...", TASK.BUSY, "Bosch Solution Plugin", g_taskHandle)
	if getInfos(lug_device) then
		getPackageVersion()
		generateData()
		if flagPasscode then
			g_appendPtr = luup.chdev.start(lul_device)
			appendAreas()
			--appendPoints()
			appendOutputs()
			luup.chdev.sync(lul_device, g_appendPtr)
			getChildDevices(lug_device)
			updateState()
			getHistoryIndexLimits()
			setHistoryInterval()
			--getHistory(3, 640)
			--[[
			StatusPoint = getPointState("11")

			log("******AREAS********")
			for k,v in pairs (g_areas) do
				log(v.panelID .. " -> " .. v.name .. " -> " .. v.devId)
			end
			log("******POINTS********")
			for k,v in pairs (g_points) do
				log(v.panelID .. " -> " .. v.name)
			end
			log("******DOORS********")
			for k,v in pairs (g_doors) do
				log(v)
			end

			log("******OUTPUTS********")
			for k,v in pairs (g_outputs) do
				log("outputs[" .. k .."] : ")
				log(v.devId)
				log(v.panelID)
				log(v.status)
				log(v.name)
			end
			-- ]]
			--peer:close()
			debug("(Startup) : Startup Successful" )
			g_taskHandle = luup.task("Startup successful.", TASK.BUSY, "Bosch Solution Plugin", g_taskHandle)
			luup.set_failure(0, lul_device)
			return true, "Startup successful.", "Bosch Security System"
		else
			peer:close()
			debug("(connectToPanel) : ERROR : Failed to connect with alarm panel, see error message above!")
			return false
		end
	else
		debug("(Startup) : Startup failed!" )
		g_taskHandle = luup.task("Please insert IP Address!", TASK.ERROR, "Bosch Solution Plugin", g_taskHandle)
		return false, "Please insert IP Address!", "Bosch Security System"
	end
end
